"""
------------------------------------------
File name: constant.py
Author: Clare Huang
"""
P_GROUND = 1000.  # Ground pressure level. Unit: hPa
SCALE_HEIGHT = 7000.  # Unit: m
CP = 1004.  # specific heat at constant pressure for air (cp) = 1004 J/kg-K
DRY_GAS_CONSTANT = 287.
EARTH_RADIUS = 6.378e+6  # Unit: m
EARTH_OMEGA = 7.29e-5
